pub(super) mod dnsupdate;
pub(super) mod pdns;

use crate::{
    macros::DefDer,
    types::{
        dns::{
            dnsupdate::{
                DNSUpdateClientOptions,
                DnsUpdateHandler,
            },
            pdns::{
                PdnsClientOptions,
                PdnsHandler,
            },
        },
        structs::Error,
        traits::{
            DnsHandler,
            DnsToken,
        },
    },
};
use log::*;
use macro_rules_attribute::macro_rules_derive;
use schemars::JsonSchema;
use serde::Deserialize;
use std::{
    collections::HashMap,
    pin::Pin,
    sync::Arc,
};
use tokio::sync::Mutex;

#[macro_rules_derive(DefDer)]
pub struct Manager(Arc<Mutex<InnerManager>>);

impl Manager {
    pub fn new() -> Self {
        Self(Arc::new(Mutex::new(InnerManager {
            servers: HashMap::new(),
        })))
    }

    pub async fn set_record(&self, domain: String, value: String) -> Option<Pin<Box<dyn DnsToken + '_>>> {
        let mut tld = domain.clone();
        if !tld.ends_with('.') {
            tld.push('.');
        }
        let (mut best_match_domain, mut best_match_length) = ("", 0);
        let guard = self.0.lock().await;
        for domain in guard.servers.keys() {
            if domain.ends_with(&tld) {
                let matched = domain.rmatches(&tld).last().unwrap();
                if matched.len() > best_match_length {
                    best_match_domain = matched;
                    best_match_length = matched.len();
                }
            }
        }
        if best_match_length == 0 {
            return None;
        }
        let handler = guard.servers.get(best_match_domain).unwrap();
        match handler.set_record(domain, value).await {
            Ok(token) => Some(token),
            Err(error) => {
                error!("Failed to set the DNS Record in the backend: {error}");
                None
            },
        }
    }
}

unsafe impl Send for Manager {}

#[derive(Debug)]
struct InnerManager {
    servers: HashMap<String, Dns>,
}


#[macro_rules_derive(DefDer)]
#[derive(Deserialize, Default, JsonSchema)]
#[serde(deny_unknown_fields)]
#[serde(tag = "type", rename_all = "lowercase")]
pub enum DnsBuilder {
    PowerDNS(PdnsClientOptions),
    DNSUpdate(DNSUpdateClientOptions),
    #[default]
    None,
}

#[macro_rules_derive(DefDer)]
enum Dns {
    PowerDNS(PdnsHandler),
    DNSUpdate(DnsUpdateHandler),
    None,
}

impl Dns {
    pub async fn set_record(&self, domain: String, content: String) -> crate::types::BoxedResult<dyn DnsToken> {
        match self {
            Dns::PowerDNS(pdns_handler) => pdns_handler.set_record(domain, content).await,
            Dns::DNSUpdate(dns_update_handler) => dns_update_handler.set_record(domain, content).await,
            Dns::None => Error::err("Not Implemented"),
        }
    }
}
