use std::{
    pin::Pin,
    time::{
        SystemTime,
        UNIX_EPOCH,
    },
};

use derive_new::new;
use log::*;
use macro_rules_attribute::macro_rules_derive;
use reqwest::{
    RequestBuilder,
    StatusCode,
};
use schemars::JsonSchema;
use serde::{
    Deserialize,
    Serialize,
};
use tokio::runtime::Handle;

use crate::{
    macros::DefDer,
    types::{
        structs::Error,
        traits::{
            DnsHandler,
            DnsToken,
        },
    },
};

#[macro_rules_derive(DefDer)]
#[derive(Deserialize, JsonSchema)]
#[serde(deny_unknown_fields)]
pub struct PdnsClientOptions {
    api_key: String,
    server: String,
    #[serde(default = "PdnsClientOptions::default_serverid")]
    server_id: String,
}


impl PdnsClientOptions {
    fn default_serverid() -> String {
        "localhost".into()
    }
}

#[macro_rules_derive(DefDer)]
pub(super) struct PdnsHandler {
    client: reqwest::Client,
    server: String,
    api_key: String,
    server_id: String,
    zone: String,
}

impl DnsHandler for PdnsHandler {
    async fn set_record(&self, domain: String, content: String) -> crate::types::BoxedResult<dyn DnsToken> {
        let base_request = self
            .client
            .patch(format!("https://{}/api/v1/servers/{}/zones/{}", self.server, self.server_id, self.zone))
            .header("Content-Type", "application/json")
            .header("X-API-Key", self.api_key.clone());

        match base_request
            .try_clone()
            .unwrap()
            .json(&RecordUpdate::new(vec![
                RRSet::new(
                    domain,
                    "TXT",
                    ChangeType::Replace {
                        records: vec![Record::new(content)],
                        comments: vec![Comment::new("ACME entry", "")],
                    },
                ),
            ]))
            .send()
            .await
        {
            Ok(_resp) if _resp.status() == StatusCode::NO_CONTENT => {},
            Ok(resp) => {
                let err = match resp.json::<PdnsError>().await {
                    Ok(error) => error.error,
                    Err(error) => format!("{error}"),
                };
                return Error::err(format!("Failed to set the record: {err}"));
            },
            Err(error) => return Error::err(format!("Failed to send the request to update the record: {error}")),
        }

        Ok(Box::pin(PdnsToken {
            builder: base_request.try_clone().unwrap().json(&0),
        }))
    }
}

/// Token that deletes the Record when its no longer needed
#[derive(Debug)]
pub(super) struct PdnsToken {
    builder: RequestBuilder,
}

impl PdnsToken {
    fn new(builder: RequestBuilder) -> Pin<Box<Self>> {
        Box::pin(Self {
            builder,
        })
    }
}

impl Drop for PdnsToken {
    fn drop(&mut self) {
        let handle = match Handle::try_current() {
            Ok(handle) => handle,
            Err(error) => {
                error!("Failed to aquire the handle from tokio. Cleanup of the Token Failed: {error}");
                return;
            },
        };
        match handle.block_on(self.builder.try_clone().unwrap().send()) {
            Ok(response) => {
                if response.status() != StatusCode::NO_CONTENT {
                    let status = response.status();
                    match handle.block_on(response.json::<PdnsError>()) {
                        Ok(error) => error!("Failed to delete the Record({status}): {error}"),
                        Err(error) => error!("Failed to parse the Error Response({status}): {error}"),
                    }
                }
            },
            Err(error) => error!("Failed to delete the Record: {error}"),
        };
    }
}

unsafe impl Send for PdnsToken {}
impl DnsToken for PdnsToken {}

#[macro_rules_derive(DefDer)]
#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub struct PdnsError {
    pub error: String,
    #[serde(default)]
    pub errors: Vec<String>,
}

#[macro_rules_derive(DefDer)]
#[derive(Serialize, new)]
struct Comment {
    #[new(into)]
    content: String,
    #[new(into)]
    account: String,

    #[new(value = "Comment::unix_epoch()")]
    modified_at: u64,
}

impl Comment {
    #[inline]
    fn unix_epoch() -> u64 {
        SystemTime::now().duration_since(UNIX_EPOCH).unwrap_or_default().as_secs()
    }
}

#[macro_rules_derive(DefDer)]
#[derive(Serialize, new)]
struct Record {
    content: String,
    #[new(value = "false")]
    disabled: bool,
}


#[macro_rules_derive(DefDer)]
#[derive(Serialize)]
#[serde(tag = "changetype", rename_all = "UPPERCASE")]
enum ChangeType {
    Replace {
        records: Vec<Record>,
        comments: Vec<Comment>,
    },
    Delete,
}

#[macro_rules_derive(DefDer)]
#[derive(Serialize, new)]
struct RRSet {
    name: String,
    #[new(into)]
    r#type: String,
    #[serde(flatten)]
    changetype: ChangeType,
}


#[macro_rules_derive(DefDer)]
#[derive(Serialize, new)]
struct RecordUpdate {
    rrset: Vec<RRSet>,
}
