use std::{
    fmt::Display,
    net::IpAddr,
};

use acme2_eab::Identifier;
use openssl::x509::GeneralName;

use crate::types::{
    config::GeneralConfig,
    dns::pdns::PdnsError,
    structs::{
        Error,
        San,
    },
};


impl Default for GeneralConfig {
    #[inline]
    fn default() -> Self {
        Self {
            accounts_path: Self::default_accounts(),
            sites_path: Self::default_sites(),
            http_challenge_path: Self::default_challenge(),
            dns: Self::default_dns(),
            certificates_path: Self::default_certificates(),
            ca: Self::default_cas(),
        }
    }
}

impl From<GeneralName> for San {
    fn from(value: GeneralName) -> Self {
        if let Some(dns) = value.dnsname() {
            return Self::Dns(dns.to_owned());
        }
        if let Some(ipaddr) = value.ipaddress() {
            if ipaddr.len() == 4 {
                let mut addr = [0u8; 4];
                addr.copy_from_slice(ipaddr);
                return Self::IPAddress(IpAddr::from(addr));
            } else {
                let mut addr = [0u8; 16];
                addr.copy_from_slice(ipaddr);
                return Self::IPAddress(IpAddr::from(addr));
            }
        }
        if let Some(email) = value.email() {
            return Self::Email(email.to_owned());
        }
        unreachable!();
    }
}
impl From<San> for Identifier {
    fn from(value: San) -> Self {
        match value {
            San::Dns(domain) => {
                Identifier {
                    r#type: "dns".into(),
                    value: domain,
                }
            },
            San::Email(email) => {
                Identifier {
                    r#type: "email".into(),
                    value: email,
                }
            },
            San::IPAddress(ip) => {
                Identifier {
                    r#type: "ip".into(),
                    value: ip.to_string(),
                }
            },
        }
    }
}

impl Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(&self.message)
    }
}


impl Display for PdnsError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(&self.error)?;
        if !self.errors.is_empty() {
            f.write_str("(")?;
            let mut iter = self.errors.iter();
            let mut error = iter.next().unwrap();
            loop {
                f.write_str(error)?;
                match iter.next() {
                    Some(err) => {
                        error = err;
                        f.write_str(", ")?;
                    },
                    None => break,
                }
            }
            f.write_str(")")?;
        }
        Ok(())
    }
}
