use std::ffi::CString;

use crate::{
    consts::{
        BRAINPOOL_MIDDLE,
        BRAINPOOL_STRONG,
        BRAINPOOL_WEAK,
        SECP_MIDDLE,
        SECP_STRONG,
        SECP_WEAK,
    },
    types::{
        cryptography::{
            Algorithm,
            Strength,
        },
        structs::Error,
    },
};
use caps::{
    CapSet,
    has_cap,
    read,
};
use log::*;
use openssl::{
    ec::EcKey,
    error::ErrorStack,
    nid::Nid,
    pkey::{
        PKey,
        Private,
    },
    rsa::Rsa,
    x509::{
        X509Name,
        X509NameBuilder,
    },
};

const CAPABILITY_SET: CapSet = CapSet::Permitted;

pub fn prefix_emails(input: Vec<String>) -> Vec<String> {
    let mut output = Vec::with_capacity(input.len());
    for mut addr in input {
        if addr.starts_with("mailto:") {
            output.push(addr);
        } else {
            addr.insert_str(0, "mailto:");
            output.push(addr);
        }
    }
    output
}


fn gen_ec_key(algorithm: Algorithm, strength: Strength) -> Result<PKey<Private>, ErrorStack> {
    let (weak, middle, strong) = match algorithm {
        Algorithm::Rsa | Algorithm::ED25519 => unreachable!(),
        Algorithm::Secp => (SECP_WEAK, SECP_MIDDLE, SECP_STRONG),
        Algorithm::Brainpool => (BRAINPOOL_WEAK, BRAINPOOL_MIDDLE, BRAINPOOL_STRONG),
    };
    let algo = EcKey::from_curve_name(match strength {
        Strength::Weak => weak,
        Strength::Middle => middle,
        Strength::Strong => strong,
    });
    let key = match algo {
        Err(error) => {
            error!("Failed to generate key due of an problem with the algorithms: {error}");
            return Err(error);
        },
        Ok(algo) => EcKey::generate(algo.group()),
    };
    match key {
        Ok(private) => PKey::from_ec_key(private),
        Err(error) => {
            error!("Failed to generate Private key from EcKey: {error}");
            Err(error)
        },
    }
}


pub fn gen_key(algorithm: Algorithm, strength: Strength) -> Result<PKey<Private>, Error> {
    let key = match algorithm {
        Algorithm::Rsa => {
            let key = Rsa::generate(strength.rsabits());
            match key {
                Ok(key) => PKey::from_rsa(key),
                Err(error) => Err(error),
            }
        },
        Algorithm::Secp => gen_ec_key(Algorithm::Secp, strength),
        Algorithm::ED25519 => PKey::generate_ed25519(),
        Algorithm::Brainpool => gen_ec_key(Algorithm::Brainpool, strength),
    };
    match key {
        Ok(key) => Ok(key),
        Err(error) => Error::err(format!("Failed to generate an key to the parameters: {error}")),
    }
}

pub fn string_to_cn(name: String) -> X509Name {
    let mut builder = X509NameBuilder::new().unwrap();
    builder.append_entry_by_nid(Nid::COMMONNAME, &name).unwrap();
    builder.build()
}

pub fn check_permissions() -> bool {
    let caps = has_cap(None, CAPABILITY_SET, caps::Capability::CAP_CHOWN).unwrap_or(false) &&
        has_cap(None, CAPABILITY_SET, caps::Capability::CAP_FOWNER).unwrap_or(false);
    if cfg!(debug_assertions) {
        for set in [
            CapSet::Ambient,
            CapSet::Bounding,
            CapSet::Effective,
            CapSet::Inheritable,
            CapSet::Permitted,
        ] {
            match read(None, set) {
                Ok(current) => {
                    trace!("Current {set:?} Set: {current:?}")
                },
                Err(error) => {
                    trace!("Failed to get the current {set:?} Set: {error}")
                },
            }
        }
    }
    caps
}

pub fn get_uid_gid(username: String, group: String) -> (Option<u32>, Option<u32>) {
    let uid = if username.contains('\0') {
        error!("Invalid Username");
        None
    } else {
        unsafe {
            match CString::new(username) {
                Ok(cstr) => {
                    let passwd = libc::getpwnam(cstr.as_ptr());
                    if !passwd.is_null() {
                        Some(passwd.read().pw_uid)
                    } else {
                        warn!("Failed to get user: User does not exist");
                        None
                    }
                },
                Err(error) => {
                    error!("Failed to convert the Username to an uid: {error}");
                    None
                },
            }
        }
    };
    let gid = if group.contains('\0') {
        error!("Invalid Group name");
        None
    } else {
        unsafe {
            match CString::new(group) {
                Ok(cstr) => {
                    let passwd = libc::getgrnam(cstr.as_ptr());
                    if !passwd.is_null() {
                        Some(passwd.read().gr_gid)
                    } else {
                        warn!("Failed to get user: User does not exist");
                        None
                    }
                },
                Err(error) => {
                    error!("Failed to convert the Group name to an uid: {error}");
                    None
                },
            }
        }
    };
    (uid, gid)
}

pub mod schema {

    use schemars::Schema;
    use serde_json::{
        Map,
        Value,
    };

    fn is_array(val: &Value) -> bool {
        if let Some(typ) = val.as_str() {
            typ == "array"
        } else if let Some(typ) = val.as_array() {
            for typvariant in typ {
                if is_array(typvariant) {
                    return true;
                }
            }
            false
        } else {
            false
        }
    }

    pub fn type_transform(schema: &mut Schema, typ: &'static str) {
        let typval = Value::String(typ.into());
        if let Some(val) = schema.get("type") {
            if is_array(val) {
                if let Some(items) = schema.get("items") {
                    let mut items = items.to_owned();
                    items.as_object_mut().unwrap().insert("format".to_string(), typval);
                    schema.insert("items".to_owned(), items);
                } else {
                    schema.insert(
                        "items".to_owned(),
                        Value::Object({
                            let mut items = Map::with_capacity(1);
                            items.insert("format".to_owned(), typval);
                            items
                        }),
                    );
                }
                return;
            }
        }
        schema.insert("format".to_string(), typval);
    }

    pub fn email_transform(schema: &mut Schema) {
        type_transform(schema, "email");
    }

    pub fn uri_transform(schema: &mut Schema) {
        type_transform(schema, "uri");
    }
}
